@@ -22,7 +22,7 @@ from typing import (
 import attr
 import bitmath
 
-from .determ_hash import determ_hash, hashable
+from .determ_hash import HASH_BYTES, determ_hash, hashable
 from .index import Index, IndexKeyType
 from .obj_store import DirObjStore, ObjStore
 from .pickler import Pickler
@@ -30,17 +30,16 @@ from .replacement_policies import REPLACEMENT_POLICIES, Entry, ReplacementPolicy
 from .rw_lock import FileRWLock, Lock, RWLock
 from .util import (
     Constant,
-    DontPickle,
     FuncParams,
     FuncReturn,
     Future,
     GetAttr,
-    KeyGen,
+    ellipsize,
     identity,
     none_tuple,
 )
 
-__version__ = "1.1.0"
+__version__ = "1.2.1"
 
 BYTE_ORDER: str = "big"
 
@@ -67,15 +66,25 @@ def memoize(
 DEFAULT_LOCK_PATH = ".cache/.lock"
 DEFAULT_OBJ_STORE_PATH = ".cache"
 
+
+def key2str(key: Tuple[Any, ...]) -> str:
+    system_key_rest = f" {key[0][1:]}" if len(key[0]) > 1 else ""
+    system_key = f"v{key[0][0]}{system_key_rest}"
+    function = f"{ellipsize(key[1], 20)} ver {determ_hash(key[2]):0{2*HASH_BYTES}x}"
+    args = f"{determ_hash(key[3]):0{2*HASH_BYTES}x} ver {determ_hash(key[4]):0{2*HASH_BYTES}x}"
+    return f"(system={system_key}, function={function}, args={args})"
+
+
 # pyright thinks attrs has ambiguous overload
 @attr.define(init=False)  # type: ignore
 class MemoizedGroup:
     """A MemoizedGroup holds the memoization for multiple functions."""
 
+    # pylint: disable=too-many-instance-attributes
+
     _index: Index[Any, Entry]
     _obj_store: ObjStore
     _replacement_policy: ReplacementPolicy
-    _key_gen: KeyGen
     _size: bitmath.Bitmath
     _pickler: Pickler
     _index_lock: RWLock
@@ -88,12 +97,21 @@ class MemoizedGroup:
     time_cost: dict[str, datetime.timedelta]
     time_saved: dict[str, datetime.timedelta]
     temporary: bool
+    logger: logging.Logger
+
+    @classmethod
+    def __determ_hash__(cls) -> Any:
+        # I don't want to put anything about the actual state of Memoized here.
+        # Because then if a Memoized function,f, called a Memoized function, g, then f will think g changed when only the memoization state chaned.
+        # If the memoization is completely transparent, then the memoization state does not affect the behavior of the program/
+        return ()
 
     def __getstate__(self) -> Any:
         return {
             slot: getattr(self, slot)
             for slot in self.__slots__
-            if slot not in {"__weakref__", "_index", "_memory_lock"}
+            if slot
+            not in {"__weakref__", "_index", "_memory_lock", "_version", "logger"}
         }
 
     def __setstate__(self, state: Mapping[str, Any]) -> Any:
@@ -109,7 +127,11 @@ class MemoizedGroup:
             ),
             self._deleter,
         )
+        self._version = 0
         self._memory_lock = threading.RLock()
+        self.logger = logging.getLogger("charmonium.cache")
+        self.logger.setLevel(logging.DEBUG)
+        self._index_read()
 
     def __init__(
         self,
@@ -134,26 +156,16 @@ class MemoizedGroup:
         :param fine_grain_persistence: De/serialize the index at every access. This is useful if you need to update the cache for multiple simultaneous processes, but it compromises performance in the single-process case.
         :param fine_grain_eviction: Maintain the cache's size through eviction at every access (rather than just the de/serialization points). This is useful if the caches size would not otherwise fit in memory, but it compromises performance if not needed.
         :param extra_system_state: A callable that returns "extra" system state. If the system state changes, the cache is dumped.
+        :param temporary: Whether the cache should be cleared at the end of the process; This is useful for tests.
 
         .. _`bitmath.Bitmath`: https://pypi.org/project/bitmath/
 
         """
-        self._index = Index[Any, Entry](
-            (
-                IndexKeyType.MATCH,  # system state
-                IndexKeyType.LOOKUP,  # func name
-                IndexKeyType.MATCH,  # func state
-                IndexKeyType.LOOKUP,  # args key
-                IndexKeyType.MATCH,  # args version
-            ),
-            self._deleter,
-        )
         self._obj_store = (
             obj_store
             if obj_store is not None
             else DirObjStore(path=DEFAULT_OBJ_STORE_PATH)
         )
-        self._key_gen = KeyGen()
         self._replacement_policy = (
             REPLACEMENT_POLICIES[replacement_policy.lower()]()
             if isinstance(replacement_policy, str)
@@ -168,69 +180,68 @@ class MemoizedGroup:
         )
         self._pickler = pickler
         self._index_lock = lock if lock is not None else FileRWLock(DEFAULT_LOCK_PATH)
-        self._memory_lock = threading.RLock()
         self._fine_grain_persistence = fine_grain_persistence
         self._fine_grain_eviction = fine_grain_eviction
         self._extra_system_state = extra_system_state
         self._index_key = 0
-        self._version = 0
         self.time_cost = DefaultDict[str, datetime.timedelta](datetime.timedelta)
         self.time_saved = DefaultDict[str, datetime.timedelta](datetime.timedelta)
         self.temporary = temporary
-        self._index_read()
+        self.__setstate__({})
         if self.temporary:
             atexit.register(self._obj_store.clear)
             # atexit handlers are run in the opposite order they are registered.
         atexit.register(self._index_write)
+        # TODO: atexit log report
 
     def _deleter(self, item: tuple[Any, Entry]) -> None:
+        # print(f"_deleter {item[0]} {determ_hash(item[0])}")
         with self._memory_lock:
             key, entry = item
             if entry.obj_store:
-                del self._obj_store[entry.value]
+                obj_key = determ_hash(key)
+                self.logger.getChild("invalidate").debug(
+                    "Invalidating key=%s obj_key=0x%x", key2str(key), obj_key
+                )
+                del self._obj_store[obj_key]
             self._replacement_policy.invalidate(key, entry)
 
     def _index_read(self) -> None:
         with self._memory_lock, self._index_lock.reader:
-            if self._index_key in self._obj_store:
-                other_version, other_index, other_rp, other_tc, other_ts = cast(
-                    Tuple[
-                        int,
-                        Index[Any, Entry],
-                        ReplacementPolicy,
-                        DefaultDict[str, datetime.timedelta],
-                        DefaultDict[str, datetime.timedelta],
-                    ],
-                    self._pickler.loads(self._obj_store[self._index_key]),
+            self._index_read_nolock()
+
+    def _index_read_nolock(self) -> None:
+        if self._index_key in self._obj_store:
+            other_version, other_index, other_rp, other_tc, other_ts = cast(
+                Tuple[
+                    int,
+                    Index[Any, Entry],
+                    ReplacementPolicy,
+                    DefaultDict[str, datetime.timedelta],
+                    DefaultDict[str, datetime.timedelta],
+                ],
+                self._pickler.loads(self._obj_store[self._index_key]),
+            )
+            if other_version > self._version:
+                self.logger.getChild("index_read").debug(
+                    "Reading index, my_version=%d, their_verison=%d",
+                    self._version,
+                    other_version,
                 )
-                if other_version > self._version:
-                    self._version = other_version
-                    self._index.update(other_index)
-                    self._replacement_policy.update(other_rp)
-                    self.time_cost = other_tc
-                    self.time_saved = other_ts
+                self._version = other_version
+                self._index.update(other_index)
+                self._replacement_policy.update(other_rp)
+                self.time_cost = other_tc
+                self.time_saved = other_ts
 
     def _index_write(self) -> None:
         with self._memory_lock, self._index_lock.writer:
-            if self._index_key in self._obj_store:
-                other_version, other_index, other_rp, other_tc, other_ts = cast(
-                    Tuple[
-                        int,
-                        Index[Any, Entry],
-                        ReplacementPolicy,
-                        DefaultDict[str, datetime.timedelta],
-                        DefaultDict[str, datetime.timedelta],
-                    ],
-                    self._pickler.loads(self._obj_store[self._index_key]),
-                )
-                if other_version > self._version:
-                    self._version = other_version
-                    self._index.update(other_index)
-                    self._replacement_policy.update(other_rp)
-                    self.time_cost = other_tc
-                    self.time_saved = other_ts
+            self._index_read_nolock()
             self._evict()
             self._version += 1
+            self.logger.getChild("_index_write").debug(
+                "Writing index version=%d", self._version
+            )
             self._obj_store[self._index_key] = self._pickler.dumps(
                 (
                     self._version,
@@ -261,24 +272,46 @@ class MemoizedGroup:
 
             while total_size > self._size:
                 key, entry = self._replacement_policy.evict()
+                obj_key = determ_hash(key)
                 if entry.obj_store:
                     assert (
-                        entry.value in self._obj_store
+                        obj_key in self._obj_store
                     ), "Replacement policy tried to evict something that wasn't there"
-                    del self._obj_store[entry.value]
+                    self.logger.getChild("_evict").debug(
+                        "Evicting key=%s obj_key=0x%x size=%r",
+                        key2str(key),
+                        obj_key,
+                        entry.data_size.format("{value:.1f}{unit}"),
+                    )
+                    del self._obj_store[obj_key]
                 total_size -= entry.data_size
                 del self._index[key]
 
-            self.remove_orphans()
-
     def remove_orphans(self) -> None:
+        """Remove data in the objstore that are not referenced by the index.
+
+        Orphans can accumulate if there are multiple processes. They
+        might generate orphans if they crash or if there is a bug in
+        my code (yikes!). If you notice accumulation of orphans, I
+        recommend calling this function once or once-per-pipeline to
+        clean them up.
+
+        However, this can compromise performance if you do it while
+        peer processes are active. They may have a slightly different
+        index-state, and you might remove something they wanted to
+        keep. I recommend calling this before you fork off processes.
+
+        """
         with self._memory_lock:
-            found_keys = {
+            found_obj_keys = {
                 entry.value for _, entry in self._index.items() if entry.obj_store
             }
-            for key in self._obj_store:
-                if key not in found_keys:
-                    del self._obj_store[key]
+            for obj_key in self._obj_store:
+                if obj_key not in found_obj_keys:
+                    self.logger.getChild("remove_orphans").debug(
+                        "Removing obj_key=0x%x", obj_key
+                    )
+                    del self._obj_store[obj_key]
 
 
 DEFAULT_MEMOIZED_GROUP = Future[MemoizedGroup].create(
@@ -289,6 +322,7 @@ DEFAULT_MEMOIZED_GROUP = Future[MemoizedGroup].create(
 # pyright thinks attrs has ambiguous overload
 @attr.define(init=False)  # type: ignore
 class Memoized(Generic[FuncParams, FuncReturn]):
+    # pylint: disable=too-many-instance-attributes
     _func: Callable[FuncParams, FuncReturn]
 
     group: MemoizedGroup
@@ -301,16 +335,10 @@ class Memoized(Generic[FuncParams, FuncReturn]):
 
     _lossy_compression: bool
 
-    # TODO: make this accept default_group_verbose = Sentinel()
-    _verbose: bool
-
     _my_pickler: Optional[Pickler]
 
     _extra_func_state: Callable[[Callable[FuncParams, FuncReturn]], Any]
 
-    _logger: logging.Logger
-    _handler: logging.Handler
-
     def __init__(
         self,
         func: Callable[FuncParams, FuncReturn],
@@ -320,7 +348,6 @@ class Memoized(Generic[FuncParams, FuncReturn]):
         use_obj_store: bool = True,
         use_metadata_size: bool = False,
         lossy_compression: bool = True,
-        verbose: bool = True,
         pickler: Optional[Pickler] = None,
         extra_func_state: Callable[[Callable[FuncParams, FuncReturn]], Any] = Constant(None),  # type: ignore
     ) -> None:
@@ -346,17 +373,9 @@ class Memoized(Generic[FuncParams, FuncReturn]):
         self._use_obj_store = use_obj_store
         self._use_metadata_size = use_metadata_size
         self._lossy_compression = lossy_compression
-        self._verbose = verbose
         self._my_pickler = pickler
         self._extra_func_state = extra_func_state
 
-        self._logger = logging.getLogger("charmonium.cache").getChild(self.name)
-        self._logger.setLevel(logging.DEBUG)
-        self._handler = DontPickle[logging.Handler].create(
-            "__import__('logging').StreamHandler(__import__('sys').stderr)"
-        )
-        self._handler.setFormatter(logging.Formatter("%(asctime)s - %(message)s"))
-
         if not self._use_obj_store and not self._use_metadata_size:
             warnings.warn(
                 "You are not using the object store, so function return-values get "
@@ -367,10 +386,6 @@ class Memoized(Generic[FuncParams, FuncReturn]):
                 UserWarning,
             )
 
-        if self._verbose:
-            atexit.register(self.log_usage_report)
-            self.enable_logging()
-
         # functools.update_wrapper(self, self._func)
 
     def log_usage_report(self) -> None:
@@ -378,7 +393,7 @@ class Memoized(Generic[FuncParams, FuncReturn]):
             tc = self.group.time_cost[self.name]
             ts = self.group.time_saved[self.name]
             print(
-                "Caching %s: cost %.1fs, saved %.1fs, net saved %.1fs"
+                "Caching %r: cost %.1fs, saved %.1fs, net saved %.1fs"
                 % (
                     self.name,
                     tc.total_seconds(),
@@ -420,7 +435,9 @@ class Memoized(Generic[FuncParams, FuncReturn]):
             self._pickler,
             # Group is a "friend class", so pylint disable
             self.group._obj_store,  # pylint: disable=protected-access
-            GetAttr[Callable[[], Any]]()(self._func, "__version__", lambda: None)(),
+            GetAttr[Callable[[], Any]]()(
+                self._func, "__version__", lambda: None, check_callable=True
+            )(),
         ) + none_tuple(self._extra_func_state(self._func))
 
     @staticmethod
@@ -434,31 +451,25 @@ class Memoized(Generic[FuncParams, FuncReturn]):
 
     def _args2key(self, *args: FuncParams.args, **kwargs: FuncParams.kwargs) -> Any:
         return {
-            key: GetAttr[Callable[[Any], Any]]()(type(val), "__cache_key__", identity)(
-                val
-            )
+            key: GetAttr[Callable[[Any], Any]]()(
+                type(val), "__cache_key__", identity, check_callable=True
+            )(val)
             for key, val in self._combine_args(*args, **kwargs).items()
         }
 
     def _args2ver(self, *args: FuncParams.args, **kwargs: FuncParams.kwargs) -> Any:
         return {
             key: GetAttr[Callable[[Any], Any]]()(
-                type(val), "__cache_ver__", Constant(())
+                type(val), "__cache_ver__", Constant(()), check_callable=True
             )(val)
             for key, val in self._combine_args(*args, **kwargs).items()
         }
 
-    def enable_logging(self) -> None:
-        self._logger.addHandler(self._handler)
-
-    def disable_logging(self) -> None:
-        self._logger.removeHandler(self._handler)
-
     def __str__(self) -> str:
         return f"memoized {self.name}"
 
     def _recompute(
-        self, *args: FuncParams.args, **kwargs: FuncParams.kwargs
+        self, obj_key: int, *args: FuncParams.args, **kwargs: FuncParams.kwargs
     ) -> tuple[Entry, FuncReturn]:
 
         start = datetime.datetime.now()
@@ -467,12 +478,13 @@ class Memoized(Generic[FuncParams, FuncReturn]):
         mid = datetime.datetime.now()
 
         if self._use_obj_store:
+            stored_value = None
             value_ser = self._pickler.dumps(value)
             data_size = bitmath.Byte(len(value_ser))
             # Group is a "friend class", hence pylint disable
-            stored_value = next(self.group._key_gen)  # pylint: disable=protected-access
+
             self.group._obj_store[  # pylint: disable=protected-access
-                stored_value
+                obj_key
             ] = value_ser
         else:
             stored_value = value
@@ -487,8 +499,8 @@ class Memoized(Generic[FuncParams, FuncReturn]):
             # pyright doesn't know attrs __init__, hence type ignore
             Entry(  # type: ignore
                 data_size=data_size,
-                recompute_time=stop - start,
-                time_saved=mid - start,
+                function_time=mid - start,
+                serialization_time=stop - mid,
                 value=stored_value,
                 obj_store=self._use_obj_store,
             ),
@@ -498,43 +510,59 @@ class Memoized(Generic[FuncParams, FuncReturn]):
     def __call__(
         self, *args: FuncParams.args, **kwargs: FuncParams.kwargs
     ) -> FuncReturn:
-        start = datetime.datetime.now()
+        call_start = datetime.datetime.now()
 
-        key = self.key(*args, **kwargs)
+        would_hit, key, obj_key = self._would_hit(*args, **kwargs)
 
-        if self.group._fine_grain_persistence:
-            with self.group._memory_lock:
-                self.group._index_read()
+        if would_hit:
+            self.group.logger.getChild("hit").debug(
+                "Hit name=%r key=%s obj_key=0x%x args_kwargs=%r",
+                self.name,
+                key2str(key),
+                obj_key,
+                ellipsize(str(args) + " " + str(kwargs), 60),
+            )
 
-        time_cost_inevitable = datetime.timedelta()
-        if key in self.group._index:
-            self._logger.debug("%s hit %s, %s", self.name, args, kwargs)
+            # Do the lookup
             with self.group._memory_lock:
                 entry = self.group._index[key]
-                if entry.obj_store:
-                    value = cast(
-                        FuncReturn,
-                        self._pickler.loads(
-                            self.group._obj_store[cast(int, entry.value)]
-                        ),
-                    )
-                else:
-                    value = cast(FuncReturn, entry.value)
-                self.group.time_saved[self.name] += entry.time_saved
+                value_ser = self.group._obj_store[obj_key] if entry.obj_store else None
+
+            # Deserialize
+            if entry.obj_store:
+                assert value_ser is not None
+                value = cast(FuncReturn, self._pickler.loads(value_ser))
+            else:
+                value = cast(FuncReturn, entry.value)
+
+            # Update time_saved
+            with self.group._memory_lock:
+                self.group.time_saved[self.name] += entry.function_time
                 self.group._replacement_policy.access(key, entry)
         else:
-            self._logger.debug("%s miss %s, %s", self.name, args, kwargs)
-            entry, value = self._recompute(*args, **kwargs)
+            self.group.logger.getChild("miss").debug(
+                "Miss name=%r key=%s obj_key=0x%x args_kwargs=%r",
+                self.name,
+                key2str(key),
+                obj_key,
+                ellipsize(str(args) + " " + str(kwargs), 60),
+            )
+
+            # Do the recompute
+            entry, value = self._recompute(obj_key, *args, **kwargs)
+
+            # Do the store
             with self.group._memory_lock:
-                time_cost_inevitable += entry.time_saved
                 if self._use_metadata_size:
-                    entry.data_size += bitmath.Byte(
-                        len(self.group._pickler.dumps(entry))
-                    )
+                    if not self._use_obj_store:
+                        entry.data_size += bitmath.Byte(
+                            len(self.group._pickler.dumps(entry))
+                        )
                     entry.data_size += bitmath.Byte(len(self.group._pickler.dumps(key)))
                 self.group._index[key] = entry
                 self.group._replacement_policy.add(key, entry)
 
+        # Update time_cost
         with self.group._memory_lock:
             if self.group._fine_grain_eviction:
                 self.group._evict()
@@ -542,8 +570,14 @@ class Memoized(Generic[FuncParams, FuncReturn]):
             if self.group._fine_grain_persistence:
                 self.group._index_write()
 
-            stop = datetime.datetime.now()
-            self.group.time_cost[self.name] += stop - start - time_cost_inevitable
+            call_stop = datetime.datetime.now()
+            time_cost_inevitable = (
+                datetime.timedelta(seconds=0) if would_hit else entry.function_time
+            )
+            # time-cost is the overhead of caching, so  it should exclud ethe overhead of the function.
+            self.group.time_cost[self.name] += (
+                call_stop - call_start - time_cost_inevitable
+            )
 
             tc = self.group.time_cost[self.name]
             ts = self.group.time_saved[self.name]
@@ -561,13 +595,18 @@ class Memoized(Generic[FuncParams, FuncReturn]):
         else:
             return obj
 
-    def key(
+    def would_hit(self, *args: FuncParams.args, **kwargs: FuncParams.kwargs) -> bool:
+        return self._would_hit(*args, **kwargs)[0]
+
+    def _would_hit(
         self, *args: FuncParams.args, **kwargs: FuncParams.kwargs
-    ) -> tuple[Any, Any, Any, Any, Any]:
+    ) -> Tuple[bool, Tuple[Any, ...], int]:
+        # pylint: disable=protected-access
+
         # Note that the system state and name or so small already, it isn't worth hashing them.
         # They are also used by other Memoized functions in the same MemoizedGroup.
         # We will only hash the potentially large key items that are used exclusively by this Memoized function.
-        return (
+        key = (
             # Group is a friend class, hence type ignore
             hashable(self.group._system_state()),  # pylint: disable=protected-access
             hashable(self.name),
@@ -575,14 +614,13 @@ class Memoized(Generic[FuncParams, FuncReturn]):
             self._compress(hashable(self._args2key(*args, **kwargs))),
             self._compress(hashable(self._args2ver(*args, **kwargs))),
         )
-
-    def would_hit(self, *args: FuncParams.args, **kwargs: FuncParams.kwargs) -> bool:
-        # pylint: disable=protected-access
-        key = self.key(*args, **kwargs)
+        obj_key = determ_hash(key)
         with self.group._memory_lock:
             if self.group._fine_grain_persistence:
                 self.group._index_read()
-            return key in self.group._index
-
-
-# TODO: work for methods (@memoize def foo(self) ...)
+            return (
+                key in self.group._index
+                and (not self._use_obj_store or obj_key in self.group._obj_store),
+                key,
+                obj_key,
+            )
